rem build.bat
rem
rem Copyright (c) 2018 The Fellowship of SML/NJ (https://smlnj.org)
rem
rem Script for SML/NJ installer to build asdlgen on Windows
rem

..\..\..\bin\ml-build.bat driver\sources.cm Main.main asdlgen

